from __future__ import annotations

import os
import shutil
from typing import Optional, Callable, Any, Dict

from yacs.config import CfgNode as ConfigurationNode
from mlmnemonist.processing_pipeline import Pipeline
from mlmnemonist.runner_cache import RunnerCache


class ExperimentRunner:
    """
    This object type is the basic object you use to run an experiment.
    It is flexible in terms of how the experiment is implemented and it is created for interactive running of experiments.
    You can separate the dataloading process from you training or testing process and you can implement a run function
    using the factory method.

    Each experiment runner has an experiment configuration file associated with it that feeds the information needed
    to conduct an experiment. You can interactively reload the configuration file mid-experiment and re-run everything.
    You can also reload the data using the load_in_dataframe option.

    As an example, checkout the 'RandomForestRuntime.ipynb' in the notebooks section to checkout an experiment runner
    that runs cv_folds on a random forest.

    This object is coupled with a FACTORY and is better to be instantiated using the factory object defined in this
    file. That way, many of the configurations will be preset to the values in the .env.
    """

    def __init__(self, experiment_dir: str, checkpoint_dir: str, verbose: int = 0,
                 cache_token: Optional[str] = None, cfg_path: str = None,
                 cfg_builder: Optional[Callable[[], ConfigurationNode]] = None) -> None:
        if cfg_path is not None and cfg_builder is not None:
            self.cfg_path = cfg_path
            self.cfg = cfg_builder()
            self.cfg.merge_from_file(self.cfg_path)

        self.verbose = verbose

        self.experiment_dir = experiment_dir
        self.checkpoint_dir = checkpoint_dir

        self._outputs: Dict[str, str] = {}

        self._implemented_run: Optional[Callable[[ExperimentRunner, ...], Any]] = None
        self.recurring_pipeline = Pipeline()
        self.preprocessing_pipeline = Pipeline()

        self.CACHE = RunnerCache(directory=checkpoint_dir, token=cache_token)

    def __str__(self) -> str:
        ret = f'Experiment runner of type: {type(self)}\n'
        ret += f'\t - cache token: {self.CACHE.TOKEN}\n'
        ret += f'\t - configurations at: {self.cfg_path}\n'
        ret += f'\t - preprocessings functions {self.preprocessing_pipeline}\n'
        ret += f'\t - recurring pipeline {self.recurring_pipeline}\n'
        if self._implemented_run is None:
            ret += f'\t - Run function not implemented!'
        else:
            ret += f'\t - Run function: {self._implemented_run.__name__}\n'
        return ret

    def reload_cfg(self) -> None:
        """
        Run this module whenever you have made a change to the cfg but you do not wish
        to redo all the previous steps on the runner. For example, when the data is loaded
        onto an experiment runner and some training configurations have changed, there is no
        need to reload the experiment runner and you can simply change the hyperparameters in
        the yaml file and reload.
        """
        self.cfg.merge_from_file(self.cfg_path)

    def preprocess(self, keep=False) -> None:
        """
        Run all the functions specified for preprocessing in an orderly fasion
        """
        self.preprocessing_pipeline.run(keep=keep, verbose=self.verbose, runner=self)

    def implement_run(self, run: Callable[[ExperimentRunner, ...], Any]) -> None:
        """
        Use this function to
        """
        self._implemented_run = run

    def ADD_OUTPUT(self, file_dir: str, description: str):
        self._outputs[file_dir] = description

    def run(self, *args, **kwargs):
        """
        This function runs an arbitrary method that is specified in the
        factory function. After running it, all the values of the config file plus a timer of the whole runtime
        will be stored in a .yaml file in the experiments directory. This will give us the possibility to re-run
        our experiments by simply specifying the output yaml file as input again.
        """
        if self._implemented_run is None:
            raise NotImplementedError("The run function is not implemented yet!")
        self.recurring_pipeline.run(keep=True, verbose=self.verbose, runner=self)
        self.CACHE._load_cache()
        ret = self._implemented_run(self, *args, **kwargs)

        # Save configurations
        if self.cfg is not None:
            name = '.'.join(os.path.basename(self.cfg_path).split('.')[:-1])
            self.cfg.dump(
                stream=open(os.path.join(self.experiment_dir, f'{name}-output.yaml'), 'w'))
            shutil.copyfile(self.cfg_path, os.path.join(self.experiment_dir, f'{name}-input.yaml'))

        # Save files
        with open(os.path.join(self.experiment_dir, 'readme.txt'), 'w') as f:
            f.writelines([f'{x}:{self._outputs[x]}' for x in self._outputs.keys()])
        return ret



import logging
import os
import textwrap
from optparse import Values
from typing import Any, List

import pip._internal.utils.filesystem as filesystem
from pip._internal.cli.base_command import Command
from pip._internal.cli.status_codes import ERROR, SUCCESS
from pip._internal.exceptions import CommandError, PipError

logger = logging.getLogger(__name__)


class CacheCommand(Command):
    """
    Inspect and manage pip's wheel cache.

    Subcommands:

    - dir: Show the cache directory.
    - info: Show information about the cache.
    - list: List filenames of packages stored in the cache.
    - remove: Remove one or more package from the cache.
    - purge: Remove all items from the cache.

    ``<pattern>`` can be a glob expression or a package name.
    """

    ignore_require_venv = True
    usage = """
        %prog dir
        %prog info
        %prog list [<pattern>] [--format=[human, abspath]]
        %prog remove <pattern>
        %prog purge
    """

    def add_options(self):
        # type: () -> None

        self.cmd_opts.add_option(
            '--format',
            action='store',
            dest='list_format',
            default="human",
            choices=('human', 'abspath'),
            help="Select the output format among: human (default) or abspath"
        )

        self.parser.insert_option_group(0, self.cmd_opts)

    def run(self, options, args):
        # type: (Values, List[Any]) -> int
        handlers = {
            "dir": self.get_cache_dir,
            "info": self.get_cache_info,
            "list": self.list_cache_items,
            "remove": self.remove_cache_items,
            "purge": self.purge_cache,
        }

        if not options.cache_dir:
            logger.error("pip cache commands can not "
                         "function since cache is disabled.")
            return ERROR

        # Determine action
        if not args or args[0] not in handlers:
            logger.error(
                "Need an action (%s) to perform.",
                ", ".join(sorted(handlers)),
            )
            return ERROR

        action = args[0]

        # Error handling happens here, not in the action-handlers.
        try:
            handlers[action](options, args[1:])
        except PipError as e:
            logger.error(e.args[0])
            return ERROR

        return SUCCESS

    def get_cache_dir(self, options, args):
        # type: (Values, List[Any]) -> None
        if args:
            raise CommandError('Too many arguments')

        logger.info(options.cache_dir)

    def get_cache_info(self, options, args):
        # type: (Values, List[Any]) -> None
        if args:
            raise CommandError('Too many arguments')

        num_http_files = len(self._find_http_files(options))
        num_packages = len(self._find_wheels(options, '*'))

        http_cache_location = self._cache_dir(options, 'http')
        wheels_cache_location = self._cache_dir(options, 'wheels')
        http_cache_size = filesystem.format_directory_size(http_cache_location)
        wheels_cache_size = filesystem.format_directory_size(
            wheels_cache_location
        )

        message = textwrap.dedent("""
            Package index page cache location: {http_cache_location}
            Package index page cache size: {http_cache_size}
            Number of HTTP files: {num_http_files}
            Wheels location: {wheels_cache_location}
            Wheels size: {wheels_cache_size}
            Number of wheels: {package_count}
        """).format(
            http_cache_location=http_cache_location,
            http_cache_size=http_cache_size,
            num_http_files=num_http_files,
            wheels_cache_location=wheels_cache_location,
            package_count=num_packages,
            wheels_cache_size=wheels_cache_size,
        ).strip()

        logger.info(message)

    def list_cache_items(self, options, args):
        # type: (Values, List[Any]) -> None
        if len(args) > 1:
            raise CommandError('Too many arguments')

        if args:
            pattern = args[0]
        else:
            pattern = '*'

        files = self._find_wheels(options, pattern)
        if options.list_format == 'human':
            self.format_for_human(files)
        else:
            self.format_for_abspath(files)

    def format_for_human(self, files):
        # type: (List[str]) -> None
        if not files:
            logger.info('Nothing cached.')
            return

        results = []
        for filename in files:
            wheel = os.path.basename(filename)
            size = filesystem.format_file_size(filename)
            results.append(f' - {wheel} ({size})')
        logger.info('Cache contents:\n')
        logger.info('\n'.join(sorted(results)))

    def format_for_abspath(self, files):
        # type: (List[str]) -> None
        if not files:
            return

        results = []
        for filename in files:
            results.append(filename)

        logger.info('\n'.join(sorted(results)))

    def remove_cache_items(self, options, args):
        # type: (Values, List[Any]) -> None
        if len(args) > 1:
            raise CommandError('Too many arguments')

        if not args:
            raise CommandError('Please provide a pattern')

        files = self._find_wheels(options, args[0])

        # Only fetch http files if no specific pattern given
        if args[0] == '*':
            files += self._find_http_files(options)

        if not files:
            raise CommandError('No matching packages')

        for filename in files:
            os.unlink(filename)
            logger.debug('Removed %s', filename)
        logger.info('Files removed: %s', len(files))

    def purge_cache(self, options, args):
        # type: (Values, List[Any]) -> None
        if args:
            raise CommandError('Too many arguments')

        return self.remove_cache_items(options, ['*'])

    def _cache_dir(self, options, subdir):
        # type: (Values, str) -> str
        return os.path.join(options.cache_dir, subdir)

    def _find_http_files(self, options):
        # type: (Values) -> List[str]
        http_dir = self._cache_dir(options, 'http')
        return filesystem.find_files(http_dir, '*')

    def _find_wheels(self, options, pattern):
        # type: (Values, str) -> List[str]
        wheel_dir = self._cache_dir(options, 'wheels')

        # The wheel filename format, as specified in PEP 427, is:
        #     {distribution}-{version}(-{build})?-{python}-{abi}-{platform}.whl
        #
        # Additionally, non-alphanumeric values in the distribution are
        # normalized to underscores (_), meaning hyphens can never occur
        # before `-{version}`.
        #
        # Given that information:
        # - If the pattern we're given contains a hyphen (-), the user is
        #   providing at least the version. Thus, we can just append `*.whl`
        #   to match the rest of it.
        # - If the pattern we're given doesn't contain a hyphen (-), the
        #   user is only providing the name. Thus, we append `-*.whl` to
        #   match the hyphen before the version, followed by anything else.
        #
        # PEP 427: https://www.python.org/dev/peps/pep-0427/
        pattern = pattern + ("*.whl" if "-" in pattern else "-*.whl")

        return filesystem.find_files(wheel_dir, pattern)
